// main.c
/********************************************************************/
/* Programm for the ADXL202-Sensor	        		    */
/* with MSP430F1232 @ 32kHz LF-XTAL                                 */                                      
/* Copyright Thomas Wedemeyer                                       */                 
/* 								    */
/* Some functions based on the MSPGCC-Examples from		    */
/* M.Mitchell (Texas Instruments, Inc) & Steve Underwood 	    */
/********************************************************************/

/*------------------------------------------------------------------------
	History:		von
	20.09.04		Thomas Wedemeyer
/*----------------------------------------------------------------------*/

/*------------------------------------------------------------------------
	Load Include-Files
/*----------------------------------------------------------------------*/
#include    <signal.h>
#include    <io.h>
#include 	<stdio.h>
#include    "define.h"

volatile unsigned char CaptureRising;
//volatile unsigned char CaptureReady = 0;

struct measurement_format
{
	volatile unsigned int StartTime;
	volatile unsigned long DutyTime;
	volatile unsigned long CycleTime;
	volatile unsigned char	Counter;
};

struct measurement_format MeasureX,MeasureXb;
struct measurement_format MeasureY,MeasureYb;
struct measurement_format MeasureZ,MeasureZb;


struct tp_format
{
	unsigned int 	min;		// -1G
	unsigned int 	zero;		// 0G
	unsigned int 	max;		// +1G 
};

struct flashmem_format 
{
	struct tp_format	x;		// x-axes
	struct tp_format	y;		// y-axes
	struct tp_format	z;		// z-axes
	unsigned int		filter;	//
	unsigned int		marker;	//
} flashmem;

struct cmd_format
{
	unsigned char		buffer[8];	// RS232-Buffer
	unsigned char		pointer;	// ArrayPointer Buffer
	unsigned char		newline;	// Newline received
} cmd;

void set_point(unsigned char type, 
				struct tp_format *axes,
				struct measurement_format * data);


/*----------------------------------------------------------------------*/
//!	Setup of the Clock LF-XTAL1
/*----------------------------------------------------------------------*/
void set_clocks()
{
    unsigned int i;

  
	WDTCTL = WDTPW + WDTHOLD;             	// Stop WDT
	BCSCTL1 |= DIVA1 + DIVA0 + RSEL2 + RSEL1 + RSEL0; // ACLK = LFXT1CLK/8
											// DCO-Resistor -> 4MHz Preset
	BCSCTL2 = 0x00;							// MCLK = DCO, DIv = 1
											// SMCLK = DCO, DIV = 1
											// DCO internal Resitor
											
	for (i = 0xFFFF; i > 0; i--);         // Delay for XTAL to settle
}  

/*----------------------------------------------------------------------*/
//!	Subroutine to set DCO to selected frequency
/*----------------------------------------------------------------------*/
void set_dco(void)                     
{
  unsigned int Compare, Oldcapture = 0;

  TACCTL2 = CCIS0 + CM0 + CAP;            // Define CCR2, CAP, ACLK
  TACTL = TASSEL1 + TACLR + MC1;        // SMCLK, continous mode
  while (1)
  {
    while ((TACCTL2 & CCIFG) != CCIFG);   // Wait until capture occured!
    TACCTL2 &= ~CCIFG;                    // Capture occured, clear flag
    Compare = TACCR2;                     // Get current captured SMCLK
    Compare = Compare - Oldcapture;     // SMCLK difference
    Oldcapture = TACCR2;                  // Save current captured SMCLK
    if (DCODELTA == Compare)
    {
      break;                            // if equal, leave "while(1)"
    }
    else if (DCODELTA < Compare)           // DCO is too fast, slow it down
    {
      DCOCTL--;
      if (DCOCTL == 0xFF)               // Did DCO role under?
      {
        BCSCTL1--;                      // Select next lower RSEL
      }
    }
    else
    {                      
      DCOCTL++;
      if (DCOCTL == 0x00)               // Did DCO role over?
      {
        BCSCTL1++;                      // Select next higher RSEL
      }
    }
  }  
  TACCTL2 = 0;                            // Stop CCR2 function
  TACTL = 0;                              // Stop Timer_A
}	

/*----------------------------------------------------------------------*/
//!	Init of USART0
/*----------------------------------------------------------------------*/
void init_usart0()
{
    U0CTL = CHAR;                         // 8-bit character
    U0TCTL = SSEL_SMCLK;                  // UCLK = SMCLK
    //U0BR0 = 0x35;                         // 2.048Mhz/38400 => 53,33
    U0BR0 = 0x6a;                         // 4.096Mhz/38400b => 71,11
    U0BR1 = 0x00;                         //
    U0MCTL = 0xa5;                        // modulation
    ME2 |= UTXE0 + URXE0;                 // Enable USART0 TXD/RXD
    IE2 |= URXIE0;                        // Enable USART0 RX interrupt
    P3SEL |= 0x30;                        // P3.4,5 = USART1 option select
    P3DIR |= 0x18;                        // P3.4, P3.3 = output direction
    P3OUT |= 0x08;                        // Enable MAX3222CA

	cmd.newline = 0;
	cmd.pointer = 0;
}


/*----------------------------------------------------------------------*/
//!	USART1 Simple putchar-function without buffer and IRQ,s
/*----------------------------------------------------------------------*/
void usart0_putchar(char character)
{
	P3OUT &= 0xFE; 							// TX-LED on

    while ((IFG2 & UTXIFG0) == 0);        // USART1 TX buffer ready?
    TXBUF0 = character;                   // RXBUF0 to TXBUF1
	
	P3OUT |= 0x01;							// TX-LED off
}

/*----------------------------------------------------------------------*/
//!	USART1 Simple putstring-function without buffer and IRQ,s
/*----------------------------------------------------------------------*/
void usart0_puts(char *character)
{
    while(*character != 0)
    {
        usart0_putchar(*character);
        character++;
    }
}

/*----------------------------------------------------------------------*/
//!	USART0 Receive Interrupt handler
/*----------------------------------------------------------------------*/
interrupt (UART0RX_VECTOR) usart0_rx(void)
{
	P3OUT &= ~0x04;
    while ((IFG2 & UTXIFG0) == 0);        // USART0 TX buffer ready?
    if(cmd.pointer < 8)
    {
    	cmd.buffer[cmd.pointer]=RXBUF0;
    	TXBUF0 = RXBUF0;                      // RXBUF0 to TXBUF0 (echo)
    	cmd.pointer++;
    }
    
    if(RXBUF0 == '\n')
    	cmd.newline = 1;
    
 	P3OUT |= 0x04;
}

/*----------------------------------------------------------------------*/
//!	Init Timer A for Capture
/*----------------------------------------------------------------------*/
void init_capture(void)                     
{

  	TACTL = TASSEL1 + TACLR + MC1;      		// SMCLK, continous mode
  												// IRQ Enable
  	TACCTL0 = CM0 + CAP + SCS + CCIE;			//RisingEdge, Cap-Mode, Sync, IRQ Enable
  	TACCTL1 = CM0 + CAP + SCS + CCIE;			//RisingEdge, Cap-Mode, Sync, IRQ Enable
  	TACCTL2 = CM0 + CAP + SCS + CCIE;			//RisingEdge, Cap-Mode, Sync, IRQ Enable
  	
  	CaptureRising = 0x07;						// Set Rising-Flags
  	P1OUT &= 0xFE;								// ADXL-Testpin = LOW
  	P2OUT &= 0xF7;								// ADXL-TestPin = Low
}

/*----------------------------------------------------------------------*/
//! Timer A0 interrupt service routine
/*----------------------------------------------------------------------*/
interrupt(TIMERA0_VECTOR) Timer_A0 (void)
{
	if(CaptureRising & 0x01)
	{
		CaptureRising &= 0x06;				// Capture on falling edge
		TACCTL0 &= 0x3FFE;
		TACCTL0 |= CM1;
	
		if(TACCR0 < MeasureY.StartTime)				// Calc CycleTime
			MeasureY.CycleTime += (unsigned long)(65536 - (MeasureY.StartTime - TACCR0));
		else
			MeasureY.CycleTime += (unsigned long)(TACCR0 - MeasureY.StartTime);
	
		MeasureY.StartTime = TACCR0;				// Save StartTime
		MeasureY.Counter++; 
	}
	else
	{
		CaptureRising |= 0x01;				// Riging Edge for Capture
		TACCTL0 &= 0x3FFE;
		TACCTL0 |= CM0;
		
		if(TACCR0 < MeasureY.StartTime)				// Calc DutyTime
			MeasureY.DutyTime += (unsigned long)(65536 -(MeasureY.StartTime - TACCR0));
		else
			MeasureY.DutyTime += (unsigned long)(TACCR0 - MeasureY.StartTime);
	}		
}

/*----------------------------------------------------------------------*/
//! Timer_A Interrupt Vector (TAIV) handler
/*----------------------------------------------------------------------*/
interrupt (TIMERA1_VECTOR) Timer_A(void)
{
  switch( TAIV )
  {
  case  2:                              // CCR1 
		 	if(CaptureRising & 0x02)
			{
				CaptureRising &= 0x05;				// Capture on falling edge
				TACCTL1 &= 0x3FFE;
				TACCTL1 |= CM1;
				
				if(TACCR1 < MeasureX.StartTime)				// Calc CycleTime
					MeasureX.CycleTime += (unsigned long)(65536 - (MeasureX.StartTime - TACCR1));
				else
					MeasureX.CycleTime += (unsigned long)(TACCR1 - MeasureX.StartTime);
				
				MeasureX.StartTime = TACCR1;				// Save StartTime
				MeasureX.Counter++;				
			}
			else
			{
				CaptureRising |= 0x02;				// Riging Edge for Capture
				TACCTL1 &= 0x3FFE;
				TACCTL1 |= CM0;
				
				if(TACCR1 < MeasureX.StartTime)				// Calc DutyTime
					MeasureX.DutyTime += (unsigned long)(65536 - (MeasureX.StartTime - TACCR1));
				else
					MeasureX.DutyTime += (unsigned long)(TACCR1 - MeasureX.StartTime);
			}	   
           break;
           
  case  4:	                       // CCR2 not used 
		 	if(CaptureRising & 0x04)
			{
				CaptureRising &= 0x03;				// Capture on falling edge
				TACCTL2 &= 0x3FFE;
				TACCTL2 |= CM1;
				
				if(TACCR2 < MeasureZ.StartTime)				// Calc CycleTime
					MeasureZ.CycleTime += (unsigned long)(65536 - (MeasureZ.StartTime - TACCR2));
				else
					MeasureZ.CycleTime += (unsigned long)(TACCR2 - MeasureZ.StartTime);
				
				MeasureZ.StartTime = TACCR2;				// Save StartTime
				MeasureZ.Counter++;				
			}
			else
			{
				CaptureRising |= 0x04;				// Riging Edge for Capture
				TACCTL2 &= 0x3FFE;
				TACCTL2 |= CM0;
				
				if(TACCR2 < MeasureZ.StartTime)				// Calc DutyTime
					MeasureZ.DutyTime += (unsigned long)(65536 - (MeasureZ.StartTime - TACCR2));
				else
					MeasureZ.DutyTime += (unsigned long)(TACCR2 - MeasureZ.StartTime);
			}	   
           break;  
  
  case 10: break;                       // overflow not used
  }
}

/*----------------------------------------------------------------------*/
//!	Init of Flash-Memory-Controller
/*----------------------------------------------------------------------*/
void init_flashcon()
{
	FCTL2 = FWKEY + FSSEL1 + FN2 + FN0;		// Setup of the Flashtiming Generator
			// Password, SMCLK / 6 => ~341kHz
}

/*----------------------------------------------------------------------*/
//!	Write Setup to FlashSegmentA 
//! @param input_poitner	Pointer to data-array
/*----------------------------------------------------------------------*/
void writeflash(struct flashmem_format *input_ptr)
{
	unsigned char *Flash_ptrA;               // Segment A pointer
	unsigned char *data_ptr;
	unsigned char i;
	unsigned char size;

	data_ptr = (unsigned char*) input_ptr;
	
	Flash_ptrA = (char *) 0x1080;   // Initialize Flash segment A pointer
	FCTL1 = FWKEY + ERASE;          // Set Erase bit
	FCTL3 = FWKEY;                  // Clear Lock bit
	*Flash_ptrA = 0;                // Dummy write to erase Flash segment B
	FCTL1 = FWKEY + WRT;            // Set WRT bit for write operation 

	size = sizeof(struct flashmem_format);
	if(size >= 128)
		size = 127;

	// Copy SegA to SegB and replace Value at the given adress
	for (i=0; i<size; i++)
  	{
   		*Flash_ptrA = *data_ptr; // copy value to segment A
   		Flash_ptrA++;
   		data_ptr++;		
	} 
  	FCTL1 = FWKEY;                  // Clear WRT bit
  	FCTL3 = FWKEY + LOCK;           // Reset LOCK bit

}

/*----------------------------------------------------------------------*/
//!	Read Int-Value from FlashSegmentA 
//! @param adr		Adressoffset in FlashSegment A
//! @return	Int-Value 
/*----------------------------------------------------------------------*/
unsigned int readflash(struct flashmem_format *output_ptr)
{
	unsigned char *Flash_ptrA;               // Segment A pointer
	unsigned char *data_ptr;
	unsigned int i;
	unsigned int size;

	Flash_ptrA = (char *) 0x1080;   // Initialize Flash segment A pointer
	data_ptr = (unsigned char *)output_ptr;

	size = sizeof(struct flashmem_format);
	if (size >= 128)
		size = 127;
	
	for(i=0; i < size; i++)
	{	
		*data_ptr = *Flash_ptrA;
   		Flash_ptrA++;
   		data_ptr++;	
	}
}


/*----------------------------------------------------------------------*/
//!	Read RS232-Buffer
/*----------------------------------------------------------------------*/
void check_buffer()
{
	unsigned char error=0;
	unsigned char itmp=0;
	
	if(cmd.buffer[0] == 'x')
	{
		
		if(cmd.buffer[1] == '+')
		{
			usart0_puts("x-DIR +\r\n");			
			set_point(DIR_P, &flashmem.x, &MeasureX);
		}
		else if(cmd.buffer[1] == '-')
		{
			usart0_puts("x-DIR -\r\n");
			set_point(DIR_N,  &flashmem.x, &MeasureX);
		}
		else if(cmd.buffer[1] == '0')
		{
			usart0_puts("x-DIR 0\r\n");
			set_point(DIR_0,  &flashmem.x, &MeasureX);
		}
		else
			error = 1;
	}

	if(cmd.buffer[0] == 'y')
	{
		usart0_puts("y-DIR\r\n");
		if(cmd.buffer[1] == '+')
			set_point(DIR_P, &flashmem.y, &MeasureY);
		else if(cmd.buffer[1] == '-')
			set_point(DIR_N, &flashmem.y, &MeasureY);
		else if(cmd.buffer[1] == '0')
			set_point(DIR_0, &flashmem.y, &MeasureY);
		else
			error = 1;
	}		
	
	if(cmd.buffer[0] == 'z')
	{
		usart0_puts("z-DIR\r\n");
		if(cmd.buffer[1] == '+')
			set_point(DIR_P, &flashmem.z, &MeasureZ);
		else if(cmd.buffer[1] == '-')
			set_point(DIR_N, &flashmem.z, &MeasureZ);
		else if(cmd.buffer[1] == '0')
			set_point(DIR_0, &flashmem.z, &MeasureZ);
		else
			error = 1;
	}
	
	if(cmd.buffer[0] == 'f')
	{
		if((cmd.buffer[1] >= '0') && (cmd.buffer[1] <= '9') &&
			(cmd.buffer[2] >= '0') && (cmd.buffer[2] <= '9'))
		{
			itmp = (cmd.buffer[1] -'0') * 10;
			flashmem.filter = itmp;
			
			itmp= cmd.buffer[2] - '0';
			flashmem.filter += itmp;
			flashmem.filter++;
			
			uprintf(usart0_putchar,"Filter = %u\r\n",flashmem.filter);
		}
		else
			error = 1;
	}
	
	if((cmd.buffer[0] == 's') && (cmd.buffer[1] == 'a'))
	{
		usart0_puts("SAVE\r\n");
		flashmem.marker = 0xAE;
		writeflash(&flashmem);
	}
	
	if(error)
	{
		usart0_puts("Error!\r\n");
	}
	
	cmd.newline = 0;
	cmd.pointer = 0;
}

/*----------------------------------------------------------------------*/
//!	Ermitteln der Abgleichpunkte
/*----------------------------------------------------------------------*/
void set_point(unsigned char type, 
				struct tp_format *axes,
				struct measurement_format * data)
{
	unsigned int duty;
	unsigned int cycle;
	unsigned int gforce;

	uprintf(usart0_putchar,"SetPoint...\r\n");
	
	while (data->Counter < 250)
	{};
	
	duty = (unsigned long)data->DutyTime / (unsigned long)data->Counter;
	cycle = (unsigned long)data->CycleTime / (unsigned long)data->Counter;
	gforce = ((unsigned long)duty * (unsigned long)10000) / (unsigned long) cycle;
	uprintf(usart0_putchar,"C=%5u  D=%5u G:%5u \r\n",cycle,duty,gforce);

	data->DutyTime = 0;
	data->CycleTime = 0;
	data->Counter = 0;

	switch(type)
	{
		case DIR_P:	// +1G
					usart0_puts("Max\r\n");
					axes->max=gforce;
					break;
		case DIR_N:	// -1G
					usart0_puts("MIN\r\n");
					axes->min=gforce;
					break;
		case DIR_0: // Zero G
					usart0_puts("Zero\r\n");
					axes->zero=gforce;
					break;
	}
}
	
/*----------------------------------------------------------------------*/
//!	Berechung der G-Force 
/*----------------------------------------------------------------------*/
int calc_force(	struct tp_format *axes, 
				struct measurement_format * data)
{
	unsigned int duty;
	unsigned int cycle;
	int force, force_p;
	
	P3OUT &= ~0x02;
	
	// Mean value duty and cycletime
	duty = (unsigned long)data->DutyTime / (unsigned long)(data->Counter);
	cycle = (unsigned long)data->CycleTime / (unsigned long)(data->Counter);
	
	// Reset filter
	data->DutyTime = 0;
	data->CycleTime = 0;
	data->Counter = 0;
	
	//Calc Duty in % * 1000
	force_p = ((unsigned long)duty * (unsigned long)10000) / (unsigned long) cycle;
	
	//Calc G-Force * 1000
	force = ((long) 2000 * (long)( (long) force_p - (long)axes->zero)) /
			((long) axes->max - (long) axes->min);

	P3OUT |= 0x02;
	return (force);
}

/*----------------------------------------------------------------------*/
//!	Send Geforce via RS232
/*----------------------------------------------------------------------*/ 
void output(int x, int y, int z)
{
	int value;
	int rest;
	
	P3OUT &= ~0x04;
	
	value = (unsigned char)(x & 0x00FF);
	usart0_putchar(value);
	
	value = (unsigned char)((x & 0xFF00 >> 8));
	usart0_putchar(value);

	value = (unsigned char)(y & 0x00FF);
	usart0_putchar(value);
	
	value = (unsigned char)((y & 0xFF00 >> 8));
	usart0_putchar(value);	
	
		value = (unsigned char)(z & 0x00FF);
	usart0_putchar(value);
	
	value = (unsigned char)((z & 0xFF00 >> 8));
	usart0_putchar(value);
	
	usart0_putchar(0xFF);
	usart0_putchar(0x00);
	usart0_putchar(0xFF);
	usart0_putchar(0x00);

	P3OUT |= 0x04;
}			
		
		
/*----------------------------------------------------------------------*/
//!	Main
/*----------------------------------------------------------------------*/
int main(void)
{
 	unsigned int i;
 	unsigned char led;
 	int xforce,yforce,zforce;
 
 	set_clocks();					// Setup Clocks
 	set_dco();						// Control-loop for DCO
	init_usart0();					// Setup of the USART
	init_flashcon();
	readflash(&flashmem);
    _EINT();						// IRQ-Enable
    
	// Setup of the i/o-pins 	
 	P1SEL |= 0x0E;					//  Capture Inputs
 	P1DIR |= 0x01;					// , P1.0 Out
 	P2DIR |= 0x08;					// P2.3 out
    P3DIR |= 0x07;                   // P3.0 .. P3.2 = Output 
    P3SEL |= 0x00;					// I/O-Port                       

	uprintf(usart0_putchar,"GForce-ADXL202\r\n");
 	
 	init_capture(); 
  
  	for(;;)
  	{

		if(cmd.newline)
		{
			check_buffer();
		}

		if((MeasureZ.Counter > flashmem.filter) && 
			(MeasureY.Counter > flashmem.filter ) &&
			(MeasureX.Counter > flashmem.filter))
		{
			xforce = calc_force( &flashmem.x, &MeasureX);
			yforce = calc_force( &flashmem.y, &MeasureY);
			zforce = calc_force( &flashmem.z, &MeasureZ);
			uprintf(usart0_putchar,"X:%i Y:%i Z:%i \r\n",xforce,yforce,zforce);  			
			//output(xforce, yforce, zforce);
		}
  	}
}

	  
	  
